# Generated by Django 4.0.7 on 2022-10-15 18:29

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("taskmonitor", "0003_queuedtask"),
    ]

    operations = [
        migrations.AddField(
            model_name="tasklog",
            name="args",
            field=models.JSONField(
                default=list,
                help_text="Positional arguments the task was called with. Nested items might be truncated to [] or {}.",
            ),
        ),
        migrations.AddField(
            model_name="tasklog",
            name="kwargs",
            field=models.JSONField(
                default=dict,
                help_text="Keyword arguments the task was called with. Nested items might be truncated to [] or {}.",
            ),
        ),
        migrations.AddField(
            model_name="tasklog",
            name="result",
            field=models.JSONField(
                default=None, help_text="Result of the task.", null=True
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="app_name",
            field=models.CharField(
                db_index=True,
                help_text="Name of the app this task belongs to.",
                max_length=255,
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="exception",
            field=models.TextField(help_text="Message of the raised exception if any."),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="parent_id",
            field=models.UUIDField(
                default=None, help_text="ID of the parent task if any.", null=True
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="priority",
            field=models.IntegerField(
                default=None,
                help_text="Priority this task was executed with.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="received",
            field=models.DateTimeField(
                default=None,
                help_text="When a task is received from the broker and is ready for execution.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="retries",
            field=models.IntegerField(help_text="Number of retries."),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="runtime",
            field=models.FloatField(
                db_index=True,
                default=None,
                help_text="Runtime of this task in seconds.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="started",
            field=models.DateTimeField(
                default=None, help_text="When the task execution started.", null=True
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="state",
            field=models.IntegerField(
                choices=[(1, "success"), (2, "retry"), (3, "failure")],
                db_index=True,
                help_text="Task's state when it was logged.",
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="task_id",
            field=models.UUIDField(
                db_index=True, default=uuid.uuid4, help_text="Unique ID of this task."
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="task_name",
            field=models.CharField(
                db_index=True, help_text="Name of this task.", max_length=255
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="timestamp",
            field=models.DateTimeField(
                db_index=True, help_text="Timestamp when this log was created."
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="traceback",
            field=models.TextField(
                help_text="Full stack trace if there was an exception."
            ),
        ),
    ]



# Generated by Django 4.0.9 on 2023-02-21 18:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("taskmonitor", "0004_store_task_params_and_results"),
    ]

    operations = [
        migrations.AddField(
            model_name="tasklog",
            name="current_queue_length",
            field=models.IntegerField(
                default=None,
                help_text="Length of the queue at the time this log was created.",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="tasklog",
            name="priority",
            field=models.IntegerField(
                db_index=True,
                default=None,
                help_text="Priority this task was executed with.",
                null=True,
            ),
        ),
    ]



import random
import string

from django import template

register = template.Library()


@register.simple_tag
def random_id(size=8) -> str:
    """Generate random id."""
    return "id-" + "".join(random.choices(string.ascii_letters + string.digits, k=size))



from celery.app import default_app

from django.core.cache import cache
from django.test import TestCase, TransactionTestCase, override_settings
from django.utils import timezone

from taskmonitor.core import task_logs, task_records
from taskmonitor.models import TaskLog

from .factories import SenderStub, TaskLogFactory

# from app_utils.testdata_factories import UserFactory


TASK_LOGS_PATH = "taskmonitor.core.task_logs"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestSignalHandlingEnd2End(TestCase):
    def setUp(self) -> None:
        cache.clear()

    def test_should_create_entry_for_succeeded_task(self):
        # given
        expected = TaskLogFactory.build(state=TaskLog.State.SUCCESS)
        task_records.set(expected.task_id, task_logs.TASK_RECEIVED, timezone.now())
        task_records.set(expected.task_id, task_logs.TASK_STARTED, timezone.now())
        sender = SenderStub.create_from_obj(expected)
        # when
        task_logs.task_success_handler_2(sender=sender, result=expected.result)
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_id=expected.task_id, state=TaskLog.State.SUCCESS
            ).exists()
        )

    def test_should_create_entry_for_failed_task(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.FAILURE, exception="", traceback=""
        )
        task_records.set(expected.task_id, task_logs.TASK_RECEIVED, timezone.now())
        task_records.set(expected.task_id, task_logs.TASK_STARTED, timezone.now())
        sender = SenderStub.create_from_obj(expected)
        other_task = TaskLogFactory.build()
        sender.request.id = str(other_task.task_id)  # now different from expected
        # when
        task_logs.task_failure_handler_2(
            sender=sender, task_id=str(expected.task_id), exception=None
        )
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_id=expected.task_id, state=TaskLog.State.FAILURE
            ).exists()
        )

    def test_should_create_entry_for_retried_task(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.RETRY, exception="", traceback=""
        )
        task_records.set(expected.task_id, task_logs.TASK_RECEIVED, timezone.now())
        task_records.set(expected.task_id, task_logs.TASK_STARTED, timezone.now())
        sender = SenderStub.create_from_obj(expected)
        sender_no_request = SenderStub.create_from_obj(expected)
        sender_no_request.request = None
        # when
        task_logs.task_retry_handler_2(
            sender=sender_no_request, request=sender.request, reason=None
        )
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_id=expected.task_id, state=TaskLog.State.RETRY
            ).exists()
        )

    def test_should_create_entry_for_internal_error_task(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.FAILURE, exception="", traceback=""
        )
        task_records.set(expected.task_id, task_logs.TASK_RECEIVED, timezone.now())
        task_records.set(expected.task_id, task_logs.TASK_STARTED, timezone.now())
        sender = SenderStub.create_from_obj(expected)
        # when
        task_logs.task_internal_error_handler_2(
            sender=sender,
            task_id=str(expected.task_id),
            request=sender.request.asdict(),
            exception=None,
        )
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_id=expected.task_id, state=TaskLog.State.FAILURE
            ).exists()
        )


@default_app.task
def normal_task():
    pass


class Dummy:
    pass


@default_app.task
def special_task():
    """Task returns an object, which is not JSON serializable."""
    return Dummy()


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestRunTasks(TransactionTestCase):
    def setUp(self) -> None:
        cache.clear()

    def test_should_create_entry_for_normal_task(self):
        # when
        normal_task.delay()
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_name="taskmonitor.tests.test_integration.normal_task",
                state=TaskLog.State.SUCCESS,
            ).exists()
        )

    def test_should_create_entry_for_task_which_result_can_not_be_serialized(self):
        # when
        special_task.delay()
        # then
        self.assertTrue(
            TaskLog.objects.filter(
                task_name="taskmonitor.tests.test_integration.special_task",
                state=TaskLog.State.SUCCESS,
            ).exists()
        )


# class TestUIEnd2End(TestCase):
#     def test_should_show_reports(self):
#         # given
#         cache.clear()
#         user = UserFactory(is_staff=True, is_superuser=True)
#         self.client.force_login(user)
#         TaskLogFactory()
#         TaskLogFactory()
#         TaskLogFactory()
#         # when
#         response = self.client.get("/taskmonitor/admin_taskmonitor_reports")
#         # then
#         self.assertEqual(response.status_code, 200)



# flake8: noqa
"""Script for creating generated task logs for testing."""

import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

import django

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

"""MAIN"""
from taskmonitor.models import TaskLog
from taskmonitor.tests.factories import TaskLogFactory

MAX_ENTRIES = 50_000

print(f"Generating {MAX_ENTRIES:,} task logs...")
objs = TaskLogFactory.build_batch(size=MAX_ENTRIES)
print("Storing...")
TaskLog.objects.bulk_create(objs, batch_size=500, ignore_conflicts=True)
print("DONE!")



from app_utils.django import clean_setting

TASKMONITOR_DATA_MAX_AGE = clean_setting("TASKMONITOR_DATA_MAX_AGE", 24)
"""Max age of logged tasks in hours. Older logs be deleted automatically."""

TASKMONITOR_HOUSEKEEPING_FREQUENCY = clean_setting(
    "TASKMONITOR_HOUSEKEEPING_FREQUENCY", 15
)
"""Frequency of house keeping runs in minutes."""

TASKMONITOR_DELETE_STALE_BATCH_SIZE = clean_setting(
    "TASKMONITOR_DELETE_STALE_BATCH_SIZE", 5_000
)
"""Size of task logs deleted together in one batch."""

TASKMONITOR_REPORTS_MAX_AGE = clean_setting("TASKMONITOR_REPORTS_MAX_AGE", 30)
"""Max age of cached reports in minutes."""

TASKMONITOR_REPORTS_MAX_TOP = clean_setting("TASKMONITOR_REPORTS_MAX_TOP", 20)
"""Max items to show in the top reports. e.g. 10 will shop the top ten items."""

TASKMONITOR_TRUNCATE_NESTED_DATA = clean_setting(
    "TASKMONITOR_TRUNCATE_NESTED_DATA", True
)
"""Whether deeply nested task params and results are truncated."""

TASKMONITOR_QUEUED_TASKS_CACHE_TIMEOUT = clean_setting(
    "TASKMONITOR_QUEUED_TASKS_CACHE_TIMEOUT", 10
)
"""Timeout for caching queued tasks in seconds. 0 will deactivate the cache."""



# Generated by Django 4.0.7 on 2022-09-06 15:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("taskmonitor", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="TaskReport",
            fields=[
                ("id", models.BigIntegerField(primary_key=True, serialize=False)),
            ],
            options={
                "verbose_name": "report",
                "managed": False,
            },
        ),
    ]



from django.apps import AppConfig

from . import __title__, __version__


class TaskMonitorConfig(AppConfig):
    name = "taskmonitor"
    label = "taskmonitor"
    default_auto_field = "django.db.models.BigAutoField"
    verbose_name = f"{__title__} v{__version__}"

    def ready(self):
        import taskmonitor.signals  # noqa: F401



from django.urls import path

from . import views

app_name = "taskmonitor"

urlpatterns = [
    path(
        "admin_taskmonitor_download_csv",
        views.admin_taskmonitor_download_csv,
        name="admin_taskmonitor_download_csv",
    ),
    path(
        "admin_taskmonitor_reports",
        views.admin_taskmonitor_reports,
        name="admin_taskmonitor_reports",
    ),
    path(
        "admin_taskmonitor_reports_clear_cache",
        views.admin_taskmonitor_reports_clear_cache,
        name="admin_taskmonitor_reports_clear_cache",
    ),
    path(
        "admin_taskmonitor_reports_recalculation",
        views.admin_taskmonitor_reports_recalculation,
        name="admin_taskmonitor_reports_recalculation",
    ),
    path(
        "admin_taskmonitor_report_json/<str:report_name>",
        views.admin_taskmonitor_report_json,
        name="admin_taskmonitor_report_json",
    ),
    path(
        "admin_taskmonitor_report_html/<str:report_name>",
        views.admin_taskmonitor_report_html,
        name="admin_taskmonitor_report_html",
    ),
    path(
        "admin_queued_task_purge",
        views.admin_queued_task_purge,
        name="admin_queued_task_purge",
    ),
    path(
        "admin_queued_task_clear_cache",
        views.admin_queued_task_clear_cache,
        name="admin_queued_task_clear_cache",
    ),
]



"""Container for caching the reports data used in views."""

import datetime as dt
import inspect
import re
import sys
from collections import defaultdict
from statistics import mean
from typing import Callable, Iterable, List, Optional, Tuple

from django.core.cache import cache
from django.db.models import Avg, Count, F, Max, Sum, Value
from django.db.models.functions import Concat, TruncMinute
from django.urls import reverse
from django.utils import functional, timezone

from ..app_settings import (
    TASKMONITOR_HOUSEKEEPING_FREQUENCY,
    TASKMONITOR_REPORTS_MAX_AGE,
    TASKMONITOR_REPORTS_MAX_TOP,
)
from ..models import TaskLog

CACHE_KEY = "taskmonitor_reports_data"
MAX_APPS_COUNT = 14
APP_NAME_OTHERS = "Others"


class _CachedReport:
    """Base class for a cached report."""

    is_included = True  # whether a report is included in the main group

    def __init__(self) -> None:
        self.name = self._to_snake_case(self.__class__.__name__)

    @staticmethod
    def _to_snake_case(name: str) -> str:
        return re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()

    @functional.cached_property
    def changelist_url(self) -> str:
        return reverse("admin:taskmonitor_tasklog_changelist")

    @functional.cached_property
    def total_runs(self) -> int:
        return TaskLog.objects.count()

    @functional.cached_property
    def total_runtime(self):
        return TaskLog.objects.aggregate(total_runtime=Sum("runtime"))["total_runtime"]

    @property
    def total_runtime_date(self):
        try:
            return self.now - dt.timedelta(seconds=self.total_runtime)
        except TypeError:
            return None

    @property
    def cache_key(self):
        return f"{CACHE_KEY}_{self.name}"

    @property
    def timeout(self):
        """Timeout in seconds."""
        return TASKMONITOR_REPORTS_MAX_AGE * 60

    @property
    def now(self) -> dt.datetime:
        return timezone.now()

    def data(self, use_cache=True) -> list:
        if use_cache:
            return cache.get_or_set(
                self.cache_key, self._calc_data, timeout=self.timeout
            )
        data = self._calc_data()
        cache.set(self.cache_key, data, timeout=self.timeout)
        return data

    def refresh_cache(self) -> None:
        """Refresh the cache."""
        cache.set(self.cache_key, self._calc_data(), timeout=self.timeout)

    def clear_cache(self) -> None:
        """Clear the cache."""
        cache.delete(self.cache_key)

    def last_update_at(self, ttl) -> Optional[dt.datetime]:
        """When the cache was last updated or None if there is no cache."""
        ttl = self._ttl()
        if not ttl:
            return None
        return timezone.now() - dt.timedelta(seconds=max(0, self.timeout - ttl))

    def next_update_at(self, ttl) -> Optional[dt.datetime]:
        """When the cache will be updated next (earliest) or None if no cache."""
        ttl = self._ttl()
        if not ttl:
            return None
        duration = TASKMONITOR_HOUSEKEEPING_FREQUENCY * 60 / 2 + ttl
        return timezone.now() + dt.timedelta(seconds=duration)

    def _ttl(self):
        return cache.ttl(self.cache_key)

    def _calc_data(self):
        """Calculate data."""
        raise NotImplementedError()

    @staticmethod
    def _truncate_minutes(
        func: Callable, qs: Iterable, minutes: int = 1
    ) -> List[Tuple[int, int]]:
        """Truncate data to one aggregated value over a span of minutes.

        Result will be sorted ascending by datetime.

        Args:
            - func: function which takes a list of values and returns one value
            - values: iterable of values in the format: [{"x": 1, "y": 2}, ...]
            - minutes: number of minutes to apply the function over.
            Must be a divider of 60.

        This function is necessary, because tests show that the ORM grouping
        with TruncMinute is not reliable.
        """

        def _func_or_zero(func, lst) -> int:
            return func(lst) if lst else 0

        if 60 % minutes > 0:
            raise ValueError("minutes must be a divider of 60.")
        data_raw = defaultdict(list)
        for x, y in qs.values_list("x", "y").iterator():
            new_minutes = x.minute // minutes * minutes
            new_x = x.replace(minute=new_minutes, second=0, microsecond=0)
            x_timestamp = int(new_x.timestamp() * 1000)
            data_raw[x_timestamp].append(y)
        data_raw = dict(sorted(data_raw.items()))
        data = [
            tuple([x, int(round(_func_or_zero(func, values), 0))])
            for x, values in data_raw.items()
        ]
        return data

    @classmethod
    def report_classes(cls):
        return [
            obj
            for _, obj in inspect.getmembers(sys.modules[__name__], inspect.isclass)
            if issubclass(obj, cls) and obj is not cls
        ]


class QueueLengthOverTime(_CachedReport):
    is_included = False

    def _calc_data(self):
        qs = (
            TaskLog.objects.exclude(current_queue_length__isnull=True)
            .order_by("timestamp")
            .annotate(x=TruncMinute("timestamp"))
            .values("x")
            .annotate(y=Avg("current_queue_length"))
        )
        data = self._truncate_minutes(mean, qs, 5)
        return [{"name": "length", "data": data}]


class TaskStatistics(_CachedReport):
    def _calc_data(self):
        qs = (
            TaskLog.objects.values("app_name", "task_name")
            .annotate(x=Count("pk"))
            .annotate(y=Avg("runtime"))
        )
        grouped_by_app = defaultdict(list)
        for obj in qs:
            point = {
                "x": obj["x"],
                "y": obj["y"],
                "task_name": obj["task_name"],
            }
            grouped_by_app[obj["app_name"]].append(point)
        data = [{"name": name, "data": data} for name, data in grouped_by_app.items()]
        return data


class TaskRunsByState(_CachedReport):
    def _calc_data(self):
        if not self.total_runs:
            return None
        return [
            {
                "name": state.label,
                "y": TaskLog.objects.filter(state=state.value).count(),
                "url": f"{self.changelist_url}?state__exact={state}",
            }
            for state in TaskLog.State
        ]


class TaskRunsByApp(_CachedReport):
    def _calc_data(self):
        if not self.total_runs:
            return None
        data = list(
            TaskLog.objects.values(name=F("app_name"))
            .annotate(y=Count("pk"))
            .annotate(url=Concat(Value(f"{self.changelist_url}?app_name="), F("name")))
            .order_by("-y")
        )
        if len(data) > MAX_APPS_COUNT:
            others_y = sum(
                [app["y"] for i, app in enumerate(data, start=1) if i > MAX_APPS_COUNT]
            )
            return data[:MAX_APPS_COUNT] + [
                {"name": APP_NAME_OTHERS, "y": others_y, "url": "#"}
            ]
        return data


class TasksTopRuns(_CachedReport):
    def _calc_data(self):
        if not self.total_runs:
            return None
        return list(
            TaskLog.objects.values(name=F("task_name"))
            .annotate(y=Count("pk"))
            .annotate(url=Concat(Value(f"{self.changelist_url}?task_name="), F("name")))
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksTopMaxRuntime(_CachedReport):
    def _calc_data(self):
        if not self.total_runtime:
            return None
        return list(
            TaskLog.objects.values(name=F("task_name"))
            .annotate(y=Max("runtime"))
            .annotate(
                url=Concat(Value(f"{self.changelist_url}?o=5&task_name="), F("name"))
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksTopAvgRuntime(_CachedReport):
    def _calc_data(self):
        if not self.total_runtime:
            return None
        return list(
            TaskLog.objects.values(name=F("task_name"))
            .annotate(y=Avg("runtime"))
            .annotate(
                url=Concat(Value(f"{self.changelist_url}?o=5&task_name="), F("name"))
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksTopFailed(_CachedReport):
    def _calc_data(self):
        total_failed = TaskLog.objects.filter(state=TaskLog.State.FAILURE).count()
        if not total_failed:
            return None
        return list(
            TaskLog.objects.filter(state=TaskLog.State.FAILURE)
            .values(name=F("task_name"))
            .annotate(y=Count("pk"))
            .annotate(
                url=Concat(
                    Value(f"{self.changelist_url}?state__exact=3&task_name="), F("name")
                )
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksTopRetried(_CachedReport):
    def _calc_data(self):
        total_retried = TaskLog.objects.filter(state=TaskLog.State.RETRY).count()
        if not total_retried:
            return None
        return list(
            TaskLog.objects.filter(state=TaskLog.State.RETRY)
            .values(name=F("task_name"))
            .annotate(y=Count("pk"))
            .annotate(
                url=Concat(
                    Value(f"{self.changelist_url}?state__exact=2&task_name="), F("name")
                )
            )
            .order_by("-y")[:TASKMONITOR_REPORTS_MAX_TOP]
        )


class TasksThroughput(_CachedReport):
    def _calc_data(self):
        tasklogs_not_failed = TaskLog.objects.exclude(state=TaskLog.State.FAILURE)
        tasks_throughput = []
        average_last_hours = dict()
        for hours in [1, 3, 6, 12, 24]:
            average_last_hours[hours] = tasklogs_not_failed.filter(
                timestamp__gt=self.now - dt.timedelta(hours=hours)
            ).avg_throughput()
        for hours, y in average_last_hours.items():
            tasks_throughput.append({"name": f"Average last {hours} hours", "y": y})
        average_overall = tasklogs_not_failed.avg_throughput()
        tasks_throughput.append({"name": "Average overall", "y": average_overall})
        peak_overall = tasklogs_not_failed.max_throughput()
        tasks_throughput.append({"name": "Peak overall", "y": peak_overall})
        return tasks_throughput


class TasksThroughputByState(_CachedReport):
    is_included = False

    def _calc_data(self):
        series = []
        for state in TaskLog.State:
            qs = (
                TaskLog.objects.filter(state=state)
                .order_by("timestamp")
                .annotate(x=TruncMinute("timestamp"))
                .values("x")
                .annotate(y=Count("id"))
            )
            data = self._truncate_minutes(sum, qs, 5)
            series.append({"name": state.label, "data": data})
        return series


class TasksThroughputByApp(_CachedReport):
    is_included = False

    def _calc_data(self):
        series = []
        app_names = [app["name"] for app in report("task_runs_by_app").data()]
        real_app_name = {name for name in app_names if name != APP_NAME_OTHERS}
        for app_name in app_names:
            if app_name in real_app_name:
                app_qs = TaskLog.objects.filter(app_name=app_name)
            else:
                app_qs = TaskLog.objects.exclude(app_name__in=real_app_name)
            qs = (
                app_qs.order_by("timestamp")
                .annotate(x=TruncMinute("timestamp"))
                .values("x")
                .annotate(y=Count("id"))
            )
            data = self._truncate_minutes(sum, qs, 5)
            series.append({"name": app_name, "data": data})
        return series


def refresh_cache() -> None:
    """Refresh the cache."""
    for report in reports():
        report.refresh_cache()


def clear_cache() -> None:
    """Clear the cache."""
    for report in reports():
        report.clear_cache()


def data() -> dict:
    """Calculate the report data."""
    return {
        report.name: report_data(report.name)
        for report in reports()
        if report.is_included
    }


def report_data(report_name: str, use_cache: bool = True):
    """Data of an cached report."""
    return report(report_name).data(use_cache)


def reports() -> List[_CachedReport]:
    """List of all cached reports."""
    return _reports.values()


def report(report_name: str) -> _CachedReport:
    """Access report by name."""
    return _reports[report_name]


# Instantiation of all cached reports
_reports = {obj.name: obj for obj in [cls() for cls in _CachedReport.report_classes()]}



from time import sleep

from django.test import TestCase

from taskmonitor.helpers import (
    compress_list,
    dict_sort_keys,
    extract_app_name,
    memcached,
    truncate_dict,
    truncate_list,
    truncate_result,
)


class TestExtractAppName(TestCase):
    def test_can_extract_from_normal_task_name(self):
        # when
        result = extract_app_name("alpha.tasks.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_can_extract_from_long_task_name(self):
        # when
        result = extract_app_name("omega.alpha.tasks.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_can_extract_from_extra_long_task_name(self):
        # when
        result = extract_app_name("echo.omega.alpha.tasks.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_can_extract_from_custom_task_name(self):
        # when
        result = extract_app_name("alpha.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_should_return_empty_string_if_no_match_1(self):
        # when
        result = extract_app_name("dummy")
        # then
        self.assertEqual(result, "")

    def test_should_return_empty_string_if_no_match_2(self):
        # when
        result = extract_app_name("tasks.dummy")
        # then
        self.assertEqual(result, "")


class TestTruncateList(TestCase):
    def test_should_copy_unnested_list(self):
        # when
        result = truncate_list([1, "alpha", 3])
        # then
        self.assertListEqual(result, [1, "alpha", 3])

    def test_should_truncate_nested_list(self):
        # when
        result = truncate_list([1, [1, 2], 3])
        # then
        self.assertListEqual(result, [1, [], 3])

    def test_should_truncate_nested_dict(self):
        # when
        result = truncate_list([1, {"alpha": 1}, 3])
        # then
        self.assertListEqual(result, [1, {}, 3])

    def test_should_truncate_tuple(self):
        # when
        result = truncate_list([1, (1, 2), 3])
        # then
        self.assertListEqual(result, [1, [], 3])

    def test_should_truncate_mix(self):
        # when
        result = truncate_list([1, [1, 2], {"alpha": 1}, (1, 2), 3])
        # then
        self.assertListEqual(result, [1, [], {}, [], 3])


class TestTruncateDict(TestCase):
    def test_should_copy_unnested_dict(self):
        # when
        result = truncate_dict({"a": 1, "b": "blue"})
        # then
        self.assertDictEqual(result, {"a": 1, "b": "blue"})

    def test_should_truncate_nested_lists(self):
        # when
        result = truncate_dict({"a": [1, 2, 3]})
        # then
        self.assertDictEqual(result, {"a": []})

    def test_should_truncate_nested_dict(self):
        # when
        result = truncate_dict({"a": {"aa": 1, "ab": 2}})
        # then
        self.assertDictEqual(result, {"a": {}})

    def test_should_truncate_mixed(self):
        # when
        result = truncate_dict({"a": 1, "b": {"ba": 1}, "c": [1, 2]})
        # then
        self.assertDictEqual(result, {"a": 1, "b": {}, "c": []})


class TestSortDict(TestCase):
    def test_should_sort_normal_keys(self):
        # when
        result = dict_sort_keys({"a": 1, "C": 3, "b": 2})
        # then
        expected = {"a": 1, "b": 2, "C": 3}
        self.assertDictEqual(result, expected)
        self.assertListEqual(list(result.keys()), list(expected.keys()))


class TestTruncateResult(TestCase):
    def test_should_copy_scalar_values(self):
        # when
        result = truncate_result("alpha")
        # then
        self.assertEqual(result, "alpha")

    def test_should_truncate_nested_lists(self):
        # when
        result = truncate_result([1, [1, 2], {"alpha": 1}, (1, 2), 3])
        # then
        self.assertListEqual(result, [1, [], {}, [], 3])

    def test_should_truncate_nested_dicts(self):
        # when
        result = truncate_result({"a": 1, "b": {"ba": 1}, "c": [1, 2]})
        # then
        self.assertDictEqual(result, {"a": 1, "b": {}, "c": []})

    def test_should_truncate_and_compress_nested_lists(self):
        # when
        result = truncate_result([[1, 2], {"alpha": 1}, (1, 2)])
        # then
        self.assertListEqual(result, [])


class TestCompressList(TestCase):
    def test_should_copy_non_empty_list_1(self):
        # when
        result = compress_list([1, 2])
        # then
        self.assertListEqual(result, [1, 2])

    def test_should_copy_non_empty_list_2(self):
        # when
        result = compress_list([1, []])
        # then
        self.assertListEqual(result, [1, []])

    def test_should_compress_list_of_empty_containers(self):
        # when
        result = compress_list([[], {}, tuple(), set()])
        # then
        self.assertListEqual(result, [])

    def test_should_keep_booleans(self):
        # when
        result = compress_list([False, []])
        # then
        self.assertListEqual(result, [False, []])


class TestMemcachedDecorator(TestCase):
    def test_should_cache_result(self):
        # given
        num_called = 0

        @memcached(timeout=1)
        def my_func():
            nonlocal num_called
            num_called += 1
            return num_called

        # when/then
        self.assertEqual(my_func(), 1)
        self.assertEqual(my_func(), 1)
        sleep(2)
        self.assertEqual(my_func(), 2)

    def test_should_cache_result_for_several_functions(self):
        # given

        @memcached()
        def my_func_1():
            return "my_func_1"

        @memcached()
        def my_func_2():
            return "my_func_2"

        # when/then
        self.assertEqual(my_func_1(), "my_func_1")
        self.assertEqual(my_func_2(), "my_func_2")



import datetime as dt
import json
import traceback as tb
from typing import List
from uuid import UUID

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Avg, Count, Max, Min
from django.db.models.functions import TruncMinute
from django.utils import timezone

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import TASKMONITOR_TRUNCATE_NESTED_DATA
from .core import celery_queues
from .helpers import extract_app_name, truncate_dict, truncate_list, truncate_result

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class QuerySetQueryStub:
    def __init__(self) -> None:
        self.select_related = None
        self.order_by = []


class ListAsQuerySet(list):
    """Masquerade a list as QuerySet."""

    def __init__(self, *args, model, distinct=False, **kwargs):
        self.model = model
        self.query = QuerySetQueryStub()
        self.distinct_enabled = distinct
        super().__init__(*args, **kwargs)
        self._id_mapper = {str(obj.id): n for n, obj in enumerate(self)}
        self._list_size = len(self)

    def all(self) -> models.QuerySet:
        return self

    def get(self, *args, **kwargs):
        try:
            return self[self._id_mapper[str(kwargs["id"])]]
        except KeyError:
            raise self.model.DoesNotExist from None

    def distinct(self):
        return ListAsQuerySet(list(set(self)), model=self.model, distinct=True)

    def values(self, *args):
        result = [
            {k: v for k, v in obj.__dict__.items() if not args or k in args}
            for obj in self
        ]
        return result

    def values_list(self, *args, **kwargs):
        items = [tuple(obj.values()) for obj in self.values(*args)]
        if kwargs.get("flat"):
            items = [obj[0] for obj in items]
            if self.distinct_enabled:
                return list(dict.fromkeys(items))
            return items
        return items

    def first(self):
        return self[0] if self else None

    def filter(self, *args, **kwargs):
        if args:
            raise NotImplementedError("filter with positional args not supported.")
        if not kwargs:
            return self
        new_list = [
            obj
            for obj in self
            if all([str(getattr(obj, k)) == str(v) for k, v in kwargs.items()])
        ]
        return ListAsQuerySet(new_list, model=self.model)

    def order_by(self, *args, **kwargs):
        if kwargs:
            raise NotImplementedError("order with kw args not supported.")
        if args:
            for prop in reversed(args):
                if prop[0:1] == "-":
                    reverse = True
                    prop = prop[1:]
                else:
                    reverse = False
                self.sort(key=lambda d: getattr(d, prop), reverse=reverse)
        return self

    def count(self):
        return self._list_size

    def _clone(self):
        return self


class QueuedTaskQuerySet(models.QuerySet):
    def count(self):
        return celery_queues.queue_length()


class QueuedTaskManagerBase(models.Manager):
    def get_queryset(self) -> models.QuerySet:
        return self.from_dto_list(celery_queues.fetch_tasks())

    def from_dto_list(self, tasks: list) -> models.QuerySet:
        """Create from a list of QueuedTaskShort objects."""
        from .models import QueuedTask

        objs = [
            QueuedTask.from_dto(obj, position) for position, obj in enumerate(tasks)
        ]
        return ListAsQuerySet(objs, model=QueuedTask)


QueuedTaskManager = QueuedTaskManagerBase.from_queryset(QueuedTaskQuerySet)


class TaskLogQuerySet(models.QuerySet):
    def csv_line_generator(self, fields: List[str]):
        """Return the tasklogs for a CSV file line by line.
        And return the field names as first line.
        """
        field_names = [field.name for field in fields]
        yield field_names
        for obj in self.iterator():
            values = [
                value for key, value in obj.asdict().items() if key in set(field_names)
            ]
            yield values

    def aggregate_timestamp_trunc(self):
        """Aggregate timestamp trunc."""
        return (
            self.annotate(timestamp_trunc=TruncMinute("timestamp"))
            .values("timestamp_trunc")
            .annotate(task_runs=Count("id"))
        )

    def max_throughput(self) -> int:
        """Calculate the maximum throughput in task executions per minute."""
        qs = self.aggregate_timestamp_trunc().aggregate(Max("task_runs"))
        return qs["task_runs__max"]

    def avg_throughput(self) -> float:
        """Calculate the average throughput in task executions per minute."""
        qs = self.aggregate_timestamp_trunc().aggregate(Avg("task_runs"))
        return qs["task_runs__avg"]

    def oldest_date(self) -> dt.datetime:
        return self.aggregate(oldest=Min("timestamp"))["oldest"]

    def newest_date(self) -> dt.datetime:
        return self.aggregate(youngest=Max("timestamp"))["youngest"]


class TaskLogManagerBase(models.Manager):
    def create_from_task(
        self,
        *,
        task_id: str,
        task_name: str,
        state: int,
        retries: int,
        priority: int,
        args: list,
        kwargs: dict,
        received: dt.datetime = None,
        started: dt.datetime = None,
        parent_id: str = None,
        exception=None,
        result=None,
        current_queue_length: int = None,
    ) -> models.Model:
        """Create new object from a celery task."""
        params = {
            "app_name": extract_app_name(task_name),
            "priority": priority,
            "parent_id": UUID(parent_id) if parent_id else None,
            "received": received,
            "retries": retries,
            "started": started,
            "state": state,
            "task_id": UUID(task_id),
            "task_name": task_name,
            "timestamp": timezone.now(),
            "current_queue_length": current_queue_length,
        }
        params["args"] = (
            truncate_list(args) if TASKMONITOR_TRUNCATE_NESTED_DATA else args
        )
        params["kwargs"] = (
            truncate_dict(kwargs) if TASKMONITOR_TRUNCATE_NESTED_DATA else kwargs
        )
        try:
            json.dumps(result, cls=DjangoJSONEncoder)
        except TypeError:
            logger.warning(
                "%s [%s]: Result was not JSON serializable and therefore discarded.",
                task_name,
                task_id,
            )
            result = None
        params["result"] = (
            truncate_result(result) if TASKMONITOR_TRUNCATE_NESTED_DATA else result
        )
        if exception:
            params["exception"] = str(exception)
            if traceback := getattr(exception, "__traceback__"):
                params["traceback"] = "".join(
                    tb.format_exception(None, value=exception, tb=traceback)
                )
        return self.create(**params)


TaskLogManager = TaskLogManagerBase.from_queryset(TaskLogQuerySet)



# flake8: noqa
"""Script for adding lots of tasks to the queue for testing."""

import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

import django

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

"""MAIN"""
from taskmonitor.core import celery_queues
from taskmonitor.tests.factories import QueuedTaskRawFactory

TASK_AMOUNT = 50_000
MAX_CHUNK_SIZE = 50_000  # upper limit to safe memory consumption


def generate_tasks(amount: int):
    tasks = (QueuedTaskRawFactory() for _ in range(amount))
    celery_queues.add_tasks(q_name, tasks)


print(f"Started adding {TASK_AMOUNT:,} tasks to queued tasks...")
q_name = celery_queues.default_queue_name()
for _ in range(TASK_AMOUNT // MAX_CHUNK_SIZE):
    generate_tasks(MAX_CHUNK_SIZE)
generate_tasks(TASK_AMOUNT % MAX_CHUNK_SIZE)
print(f"Using queue name: {q_name}")
print(f"Added {TASK_AMOUNT:,} to queued tasks.")
celery_queues.tasks_cache.clear()
print("DONE!")



import json
from collections import Counter
from typing import Optional

from django.contrib import admin
from django.db.models import Count
from django.shortcuts import get_object_or_404, redirect
from django.utils import html, safestring, timezone
from django.utils.translation import gettext_lazy as _

from .app_settings import TASKMONITOR_QUEUED_TASKS_CACHE_TIMEOUT
from .core import celery_queues
from .models import QueuedTask, TaskLog, TaskReport


class FieldFilterCountsMemory(admin.SimpleListFilter):
    """Filter by field and show counts.

    Counts are calculated in memory.
    """

    field_name = ""  # field to filter by

    def lookups(self, request, model_admin: admin.ModelAdmin):
        field_in_rows = model_admin.get_queryset(request).values_list(
            self.field_name, flat=True
        )
        field_counts = Counter(field_in_rows)
        result = [
            (field, f"{field} ({count:,})") for field, count in field_counts.items()
        ]
        return sorted(result, key=lambda obj: obj[0])

    def queryset(self, request, queryset):
        if self.value():
            params = {self.field_name: self.value()}
            return queryset.filter(**params)


class QueuedTaskAppsListFilter(FieldFilterCountsMemory):
    """Filter by app name and show name with counts."""

    title = _("app name")
    parameter_name = "app_name"
    field_name = "app_name"


class QueuedTaskTasksListFilter(FieldFilterCountsMemory):
    """Filter by task name and show name with counts."""

    title = _("task name")
    parameter_name = "task_name"
    field_name = "name"


@admin.register(QueuedTask)
class QueuedTaskAdmin(admin.ModelAdmin):
    list_display = (
        "position",
        "id",
        "name",
        "priority",
        "app_name",
    )
    list_display_links = None
    list_filter = (QueuedTaskAppsListFilter, "priority", QueuedTaskTasksListFilter)
    ordering = ["position"]

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or {}
        cache_created_at = celery_queues.tasks_cache.created_at() or timezone.now()
        context = {
            "title": "Currently queued tasks",
            "cache_created_at": cache_created_at,
            "task_count": QueuedTask.objects.count(),
            "cache_timeout": TASKMONITOR_QUEUED_TASKS_CACHE_TIMEOUT,
        }
        extra_context.update(context)
        return super().changelist_view(request, extra_context)


@admin.register(TaskReport)
class TaskReportAdmin(admin.ModelAdmin):
    def has_add_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def changelist_view(self, request, extra_context=None):
        return redirect("taskmonitor:admin_taskmonitor_reports")


class FieldFilterCountsDb(admin.SimpleListFilter):
    """Filter by field and show counts.

    Counts are calculated by the database.
    """

    field_name = ""  # field to filter by

    def lookups(self, request, model_admin: admin.ModelAdmin):
        qs = model_admin.get_queryset(request)
        field_counts = (
            qs.values(self.field_name)
            .annotate(num_words=Count(self.field_name))
            .order_by(self.field_name)
        )
        field = qs.model._meta.get_field(self.field_name)
        if field.choices:
            field_counts = self._map_choices_field(field, field_counts)
            result = [
                (
                    obj[self.field_name][0],
                    f'{obj[self.field_name][1]} ({obj["num_words"]:,})',
                )
                for obj in field_counts
            ]
        else:
            result = [
                (obj[self.field_name], f'{obj[self.field_name]} ({obj["num_words"]:,})')
                for obj in field_counts
            ]
        return result

    def _map_choices_field(self, field, field_counts):
        """Map choices field values to corresponding labels and keep values."""
        mapper = {obj[0]: obj[1] for obj in field.choices}
        field_counts = [
            {
                self.field_name: (obj[self.field_name], mapper[obj[self.field_name]]),
                "num_words": obj["num_words"],
            }
            for obj in field_counts
        ]
        return field_counts

    def queryset(self, request, queryset):
        if self.value():
            params = {self.field_name: self.value()}
            return queryset.filter(**params)


class TaskLogAppsListFilter(FieldFilterCountsDb):
    """Filter by app name and show name with counts."""

    title = _("app name")
    parameter_name = "app"
    field_name = "app_name"


class TaskLogStatesListFilter(FieldFilterCountsDb):
    """Filter by app name and show name with counts."""

    title = _("state")
    parameter_name = "state"
    field_name = "state"


class TaskLogTasksListFilter(FieldFilterCountsDb):
    """Filter by app name and show name with counts."""

    title = _("task name")
    parameter_name = "task"
    field_name = "task_name"


@admin.register(TaskLog)
class TaskLogAdmin(admin.ModelAdmin):
    class Media:
        css = {"all": ("taskmonitor/css/admin.css",)}

    list_display = (
        "timestamp",
        "task_name",
        "_params",
        "priority",
        "_state",
        "_runtime",
        "_exception",
    )
    list_filter = (
        TaskLogStatesListFilter,
        "timestamp",
        TaskLogAppsListFilter,
        "priority",
        TaskLogTasksListFilter,
    )
    search_fields = ("task_name", "app_name", "task_id")
    actions = ["delete_selected_2"]
    show_full_result_count = False
    fields = (
        "task_id",
        "task_name",
        "timestamp",
        "_args",
        "_kwargs",
        "_result",
        "retries",
        "priority",
        "state",
        "runtime",
        "app_name",
        "_exception",
        "parent_id",
        "received",
        "started",
        "_traceback",
    )

    def has_add_permission(self, *args, **kwargs) -> bool:
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def get_actions(self, request):
        actions = super().get_actions(request)
        if "delete_selected" in actions:
            del actions["delete_selected"]
        return actions

    def get_readonly_fields(self, request, obj):
        try:
            field = [f for f in obj._meta.fields if f.name == "kwargs"]
            if len(field) > 0:
                field = field[0]
                field.help_text = "some special help text"
        except Exception:
            pass
        return self.readonly_fields

    def _params(self, obj):
        if obj.args and not obj.args:
            return html.format_html("<code>{}</code>", json.dumps(obj.args))
        if not obj.args and obj.kwargs:
            return html.format_html(
                "<code>{}</code>", json.dumps(obj.kwargs, sort_keys=True)
            )
        if obj.args and obj.kwargs:
            return html.format_html(
                "<code>{}<br>{}</code>",
                json.dumps(obj.args),
                json.dumps(obj.kwargs, sort_keys=True),
            )
        return None

    @admin.display(ordering="runtime")
    def _runtime(self, obj) -> Optional[str]:
        return f"{obj.runtime:.1f}" if obj.runtime else None

    @admin.display(ordering="state")
    def _state(self, obj) -> str:
        css_class_map = {
            TaskLog.State.RETRY: "state-retry",
            TaskLog.State.FAILURE: "state-failure",
        }
        css_class = css_class_map.get(obj.state, "")
        return html.format_html(
            '<span class="{}">{}</span>', css_class, obj.get_state_display()
        )

    @admin.display(description="Exception")
    def _exception(self, obj) -> str:
        return html.format_html("<code>{}</code>", obj.exception)

    @admin.action(description="Delete selected entries (NO CONFIRMATION!")
    def delete_selected_2(self, request, queryset):
        entries_count = queryset.count()
        queryset._raw_delete(queryset.db)
        self.message_user(request, f"Deleted {entries_count} entries.")

    @admin.display(description="Result")
    def _result(self, obj):
        if obj.state == TaskLog.State.SUCCESS:
            return format_html_data(obj.result)
        return "-"

    @admin.display(description="Args")
    def _args(self, obj):
        return format_html_data(obj.args)

    @admin.display(description="Kwargs")
    def _kwargs(self, obj):
        return format_html_data(obj.kwargs)

    @admin.display(description="Traceback")
    def _traceback(self, obj):
        return format_html_lines(obj.traceback)

    def change_view(self, request, object_id, form_url="", extra_context=None):
        extra_context = extra_context or {}
        obj = get_object_or_404(TaskLog, pk=object_id)
        extra_context["tasklog_text"] = obj.astext()
        return super().change_view(
            request, object_id, form_url, extra_context=extra_context
        )


def format_html_lines(text) -> str:
    return safestring.mark_safe(
        "<br>".join(
            [html.format_html("<code>{}</code>", line) for line in text.splitlines()]
        )
    )


def format_html_data(data) -> str:
    return html.format_html(
        "<code>{}</code>", json.dumps(data, sort_keys=True, indent=4)
    )



"""Celery signal bindings.

This module is kept intentionally small, since it is difficult to test signals directly.
"""
from celery import signals

from .core import task_logs


@signals.task_received.connect
def task_received_handler(request=None, **kw):
    task_logs.task_received_handler_2(request)


@signals.task_prerun.connect
def task_prerun_handler(task_id=None, **kw):
    task_logs.task_prerun_handler_2(task_id)


@signals.task_retry.connect
def task_retry_handler(sender=None, request=None, reason=None, **kw):
    task_logs.task_retry_handler_2(sender=sender, request=request, reason=reason)


@signals.task_success.connect
def task_success_handler(sender=None, result=None, **kw):
    task_logs.task_success_handler_2(sender=sender, result=result)


@signals.task_failure.connect
def task_failure_handler(sender=None, task_id=None, exception=None, **kw):
    task_logs.task_failure_handler_2(sender, task_id, exception)


@signals.task_internal_error.connect
def task_internal_error_handler(
    sender=None, task_id=None, request=None, exception=None, **kw
):
    task_logs.task_internal_error_handler_2(sender, task_id, request, exception)



from django.core.cache import cache
from django.test import TestCase
from django.utils import timezone

from taskmonitor.core import task_records


class TestTaskRecords(TestCase):
    def setUp(self) -> None:
        cache.clear()

    def test_should_store_and_retrieve_data(self):
        # given
        value = timezone.now()
        # when
        task_records.set("abc", "alpha", value)
        result = task_records.get("abc", "alpha")
        # then
        self.assertEqual(result, value)

    def test_should_delete_data(self):
        # given
        value = timezone.now()
        # when
        task_records.set("abc", "alpha", value)
        result = task_records.fetch("abc", "alpha")
        # then
        self.assertEqual(result, value)
        self.assertIsNone(task_records.get("abc", "alpha"))



import datetime as dt

from celery import shared_task

from django.utils import timezone

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from app_utils.helpers import chunks
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import TASKMONITOR_DATA_MAX_AGE, TASKMONITOR_DELETE_STALE_BATCH_SIZE
from .core import cached_reports
from .models import TaskLog

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

DEFAULT_TASK_PRIORITY = 4


@shared_task(base=QueueOnce)
def run_housekeeping():
    """Run all housekeeping tasks."""
    delete_stale_tasklogs.apply_async(priority=7)
    refresh_reports_cache.apply_async(priority=DEFAULT_TASK_PRIORITY)


@shared_task
def delete_stale_tasklogs():
    """Delete all stale tasklogs from the database."""
    log_pks = TaskLog.objects.filter(
        timestamp__lte=timezone.now() - dt.timedelta(hours=TASKMONITOR_DATA_MAX_AGE)
    ).values_list("pk", flat=True)
    for log_pks_chunk in chunks(log_pks, TASKMONITOR_DELETE_STALE_BATCH_SIZE):
        delete_tasklogs_batch.apply_async(priority=7, kwargs={"log_pks": log_pks_chunk})


@shared_task
def delete_tasklogs_batch(log_pks: list):
    """Delete a selection of tasklogs."""
    logs_to_delete = TaskLog.objects.filter(pk__in=log_pks)
    logger.info(f"Deleting {logs_to_delete.count():,} stale tasklogs.")
    logs_to_delete._raw_delete(logs_to_delete.db)


@shared_task
def refresh_reports_cache():
    """Refresh cache for all reports."""
    reports = cached_reports.reports()
    logger.info(f"Refreshing caches for {len(reports)} reports...")
    for report in reports:
        refresh_single_report_cache.apply_async(
            priority=DEFAULT_TASK_PRIORITY, args=[report.name]
        )


@shared_task
def refresh_single_report_cache(report_name: str):
    """Refresh cache for given report."""
    cached_reports.report(report_name).refresh_cache()
    logger.info(f"Refreshed reports cache of {report_name}.")
